PROMPT_ROLE = """
Your name is: {name}.

You are designed to be able to assist with a wide range of tasks, from answering simple questions to providing in-depth explanations and discussions on a wide range of topics. As an Agent, you are able to generate human-like text based on the input it receives, allowing it to engage in natural-sounding conversations and provide responses that are coherent and relevant to the topic at hand.

You are constantly learning and improving, and its capabilities are constantly evolving. It is able to process and understand large amounts of text, and can use this knowledge to provide accurate and informative responses to a wide range of questions. Additionally, you are able to generate its own text based on the input it receives, allowing it to engage in discussions and provide explanations and descriptions on a wide range of topics.

Overall, you are a powerful Agent that can help with a wide range of tasks and provide valuable insights and information on a wide range of topics. Whether you need help with a specific question or just want to have a conversation about a particular topic, you are here to assist.

The current date and time is: {date}.
"""

PROMPT_INSTRUCTIONS = """
## Instructions:
{instructions}
"""

PROMPT_KNOWLEDGE = """
## Use the following references from the knowledge base if it helps:

<references>
{references}
<references>
"""

PROMPT_CONTEXT = """
## Use the following information if it helps:

{context}
"""

PROMPT_STEPS = """
## Before providing your final answer, follow these steps:

<steps>
{reasoning_steps}
</steps>

This structured approach will help you organize your thoughts and ensure a thorough response.
"""

PROMPT_PREVIOUS_INTERACTIONS = """
## Answer the question considering the previous interactions:

<previous_interactions>
{previous_interactions}
</previous_interactions>
"""

PROMPT_JSON_SCHEMA = """
## Provide your output using the following JSON schema:

<json_schema>
{json_schema}
</json_schema>
"""

PROMPT_REASONING_STEPS_TOOLS = """
Your objective is to thoroughly research your task using the following tools as your primary source and provide a detailed and informative answer.

You have access to the following tools:

<tools>
{tools}
</tools>

Please provide the detailed steps that are needed to achieve your task accurately and efficiently.
"""

PROMPT_REASONNING = """
I am building a reAct agent. You are the thinker of the agent.  
Instructions:
- Given <task>, provide a very simple structured approach to solve the <task>.
- Enumerate your steps in one list.
- DO NOT answer to <task>.
- You are answering to a large language model.
"""



from gwenflow.types.message import Message
from gwenflow.types.document import Document
from gwenflow.types.completion import ChatCompletion, ChatCompletionChunk, ChatCompletionMessageToolCall

__all__ = [
    "Message",
    "Document",
    "ChatCompletion",
    "ChatCompletionChunk",
    "ChatCompletionMessageToolCall",
]



from pydantic import BaseModel, ConfigDict
from typing import Any, Dict, Optional, Union, List


class Message(BaseModel):
    """Chat message class."""

    role: str
    """The role of the messages author (system, user, assistant or tool)."""

    content: Optional[Union[str, list[Union[str, dict]]]] = None
    """Content of the message."""

    name: Optional[str] = None
    """An optional name for the participant."""

    tool_call_id: Optional[str] = None
    """Tool call that this message is responding to."""
    
    tool_calls: Optional[List[Dict[str, Any]]] = None
    """The tool calls generated by the model, such as function calls."""

    thinking: Optional[str] = None
    """Thinking content from the model."""

    model_config = ConfigDict(extra="allow", populate_by_name=True, arbitrary_types_allowed=True)


    def __str__(self) -> str:
        return f"{self.role}: {self.content}"

    def to_dict(self, **kwargs: Any) -> Dict[str, Any]:
        message_dict = self.model_dump(**kwargs)
        message_dict = {
            k: v for k, v in message_dict.items() if v is not None and not (isinstance(v, (list, dict)) and len(v) == 0)
        }
        return message_dict




from typing import Optional, List
from pydantic import BaseModel


class CompletionUsage(BaseModel):
    prompt_tokens: int
    """The number of tokens used by the prompt."""

    completion_tokens: int
    """Number of tokens in the generated completion."""

    total_tokens: int
    """The total number of tokens used by the request."""

class Function(BaseModel):
    arguments: str
    name: Optional[str] = None
    """The name of the function to call."""


class ChatCompletionMessageToolCall(BaseModel):
    id: Optional[str] = None
    """The ID of the tool call."""

    function: Function
    """The function that the model called."""

    type: Optional[str] = None
    """The type of the tool. Currently, only `function` is supported."""


class ChatCompletionAudio(BaseModel):
    id: str
    """Unique identifier for this audio response."""

    data: str
    """
    Base64 encoded audio bytes generated by the model, in the format specified in
    the request.
    """

    expires_at: int
    """
    The Unix timestamp (in seconds) for when this audio response will no longer be
    accessible on the server for use in multi-turn conversations.
    """

    transcript: str
    """Transcript of the audio generated by the model."""


class ChatCompletionMessage(BaseModel):
    content: Optional[str] = None
    """The contents of the message."""

    refusal: Optional[str] = None
    """The refusal message generated by the model."""

    role: Optional[str] = None
    """The role of the author of this message."""

    audio: Optional[ChatCompletionAudio] = None
    """
    If the audio output modality is requested, this object contains data about the
    audio response from the model.
    """

    tool_calls: Optional[List[ChatCompletionMessageToolCall]] = None
    """The tool calls generated by the model, such as function calls."""


class Choice(BaseModel):
    finish_reason: Optional[str] = None
    """The reason the model stopped generating tokens."""

    index: int
    """The index of the choice in the list of choices."""

    # logprobs: Optional[ChoiceLogprobs] = None
    # """Log probability information for the choice."""

    message: ChatCompletionMessage
    """A chat completion message generated by the model."""


class ChoiceStream(BaseModel):
    finish_reason: Optional[str] = None
    """The reason the model stopped generating tokens."""

    index: int
    """The index of the choice in the list of choices."""

    # logprobs: Optional[ChoiceLogprobs] = None
    # """Log probability information for the choice."""

    delta: ChatCompletionMessage
    """A chat completion message generated by streamed model."""


class ChatCompletionBase(BaseModel):
    id: str
    """A unique identifier for the chat completion."""

    created: int
    """The Unix timestamp (in seconds) of when the chat completion was created."""

    model: str
    """The model used for the chat completion."""

    system_fingerprint: Optional[str] = None
    """Backend configuration that the model runs with."""

    usage: Optional[CompletionUsage] = None
    """Usage statistics for the completion request."""


class ChatCompletion(ChatCompletionBase):
    choices: List[Choice]
    """A list of chat completion choices."""

class ChatCompletionChunk(ChatCompletionBase):
    choices: List[ChoiceStream]
    """A list of chat completion choices."""




from typing import List, Dict, Any, Optional
from pydantic import BaseModel, model_validator, field_validator, Field

import yaml

from gwenflow.agents import Agent
from gwenflow.llms import ChatBase
from gwenflow.tools import Tool, BaseTool
from gwenflow.utils import logger


MAX_TRIALS=5

class FlowStep(BaseModel):
    agent: Agent
    task: str | None = None
    depends_on: List[str] = []
    final_step: bool = False

class Flow(BaseModel):
    steps: Optional[list] = []
    manager: Optional[Agent] = None
    llm: ChatBase
    tools: List[BaseTool] = []

    flow_type: str = "sequence"

    @model_validator(mode="after")
    def model_valid(self) -> Any:
        if self.manager is None:
            self.manager = Agent(
                name="Team Manager",
                instructions= [
                    "You are the leader of a team of AI Agents.",
                    "Manage the team to complete the task in the best way possible.",
                    "Even though you don't perform tasks by yourself, you have a lot of experience in the field, which allows you to properly evaluate the work of your team members.",
                    "You must always validate the output of the other Agents and you can re-assign the task if you are not satisfied with the result.",
                ],
                tools=self.tools,
                llm=self.llm,
            )
        return self
        
    @classmethod
    def from_yaml(cls, file: str, tools: List[Tool], llm: Optional[Any] = None) -> "Flow":
        if cls == Flow:
            with open(file) as stream:
                try:
                    agents = []
                    content_yaml = yaml.safe_load(stream)
                    for name in content_yaml.get("agents").keys():

                        _values = content_yaml["agents"][name]

                        _tools = []
                        if _values.get("tools"):
                            _agent_tools = _values.get("tools").split(",")
                            for t in tools:
                                if t.name in _agent_tools:
                                    _tools.append(t)

                        depends_on = None
                        if _values.get("depends_on"):
                            depends_on = _values.get("depends_on")                            

                        agent = Agent(
                            name=name,
                            description=_values.get("description"),
                            response_model=_values.get("response_model"),
                            tools=_tools,
                            depends_on=depends_on,
                            llm=llm,
                        )
                        agents.append(agent)
                    return Flow(agents=agents)
                except Exception as e:
                    logger.error(repr(e))
        raise NotImplementedError(f"from_yaml not implemented for {cls.__name__}")
    
    def describe(self):
        for step in self.steps:
            step = FlowStep(**step)
            print("---")
            print(f"Agent  : {step.agent.name}")
            if step.depends_on:
                print(f"Depends on:", ",".join(step.depends_on))
            if step.agent.tools:
                available_tools = [ tool.name for tool in step.agent.tools ]
                print(f"Tools  :", ",".join(available_tools))


    def run(self, query: str) -> str:

        outputs = {}

        while len(outputs) < len(self.steps):

            for step in self.steps:

                step = FlowStep(**step)

                # check if already run
                if step.agent.name in outputs.keys():
                    continue

                # check agent dependancies
                if step.depends_on:
                    context = None
                    if any(outputs.get(agent_name) is None for agent_name in step.depends_on):
                        continue
                    if step.depends_on:
                        context = { f"{agent_name}": outputs[agent_name].content for agent_name in step.depends_on }            
                    if step.task:
                        outputs[step.agent.name] = step.agent.run(step.task, context=context)
                    else:
                        outputs[step.agent.name] = step.agent.run(query, context=context)
                
                # no dependency
                else:
                    if step.task:
                        outputs[step.agent.name] = step.agent.run(step.task)
                    else:
                        outputs[step.agent.name] = step.agent.run(query)

        return outputs
    


from typing import Optional, Union, Any, List, Dict
from pydantic import BaseModel, ConfigDict
from abc import ABC, abstractmethod

import asyncio
import json

from gwenflow.tools import BaseTool
from gwenflow.types import Message, ChatCompletionMessageToolCall
from gwenflow.utils import logger


LLM_CONTEXT_WINDOW_SIZES = {
    # openai
    "gpt-4": 8192,
    "gpt-4o": 128000,
    "gpt-4o-mini": 128000,
    "gpt-4-turbo": 128000,
    "o1-preview": 128000,
    "o1-mini": 128000,
    "o3-mini": 128000,
    # deepseek
    "deepseek-chat": 128000,
    "deepseek-r1": 128000,
    # google
    "gemma2-9b-it": 8192,
    "gemma-7b-it": 8192,
    # meta
    "llama3-groq-70b-8192-tool-use-preview": 8192,
    "llama3-groq-8b-8192-tool-use-preview": 8192,
    "llama-3.1-70b-versatile": 131072,
    "llama-3.1-8b-instant": 131072,
    "llama-3.2-1b-preview": 8192,
    "llama-3.2-3b-preview": 8192,
    "llama-3.2-11b-text-preview": 8192,
    "llama-3.2-90b-text-preview": 8192,
    "llama3-70b-8192": 8192,
    "llama3-8b-8192": 8192,
    # mistral
    "mixtral-8x7b-32768": 32768,
}

class ChatBase(BaseModel, ABC):
 
    model: str
    system_prompt: Optional[str] = None
    tools: List[BaseTool] = []
    tool_choice: Optional[Union[str, Dict[str, Any]]] = None
    
    model_config = ConfigDict(arbitrary_types_allowed=True)

    @abstractmethod
    def invoke(self, *args, **kwargs) -> Any:
        pass

    @abstractmethod
    async def ainvoke(self, *args, **kwargs) -> Any:
        pass

    @abstractmethod
    def stream(self, *args, **kwargs) -> Any:
        pass
 
    @abstractmethod
    async def astream(self, *args, **kwargs) -> Any:
        pass

    @abstractmethod
    def response(self, *args, **kwargs) -> Any:
        pass

    @abstractmethod
    async def aresponse(self, *args, **kwargs) -> Any:
        pass

    @abstractmethod
    def response_stream(self, *args, **kwargs) -> Any:
        pass
 
    @abstractmethod
    async def aresponse_stream(self, *args, **kwargs) -> Any:
        pass

    def get_context_window_size(self) -> int:
        # Only using 75% of the context window size to avoid cutting the message in the middle
        return int(LLM_CONTEXT_WINDOW_SIZES.get(self.model, 8192) * 0.75)

    def _cast_messages(self, messages: Union[str, List[Message], List[Dict[str, str]]],) -> List[Message]:
        if isinstance(messages, str):
            _messages = [Message(role="user", content=messages)]
        else:
            _messages = messages
            for i, message in enumerate(_messages):
                if not isinstance(message, Message):
                    _messages[i] = Message(**message)
        return _messages

    def get_tool_names(self):
        return [tool.name for tool in self.tools]

    def get_tool_map(self):
        return {tool.name: tool for tool in self.tools}

    def handle_tool_call(self, tool_call) -> Message:

        if isinstance(tool_call, dict):
            tool_call = ChatCompletionMessageToolCall(**tool_call)
    
        tool_map  = self.get_tool_map()
        tool_name = tool_call.function.name
                    
        if tool_name not in tool_map.keys():
            logger.error(f"Tool {tool_name} does not exist")
            return Message(
                role="tool",
                tool_call_id=tool_call.id,
                tool_name=tool_name,
                content=f"Tool {tool_name} does not exist",
            )

        try:
            function_args = json.loads(tool_call.function.arguments)
        except json.JSONDecodeError as e:
            logger.error(f"Failed to parse tool arguments: {e}")
            return Message(
                role="tool",
                tool_call_id=tool_call.id,
                tool_name=tool_name,
                content=f"Failed to parse tool arguments: {e}",
            )

        try:
            logger.debug(f"Tool call: {tool_name}({function_args})")
            result = tool_map[tool_name].run(**function_args)
            if result:
                return Message(
                    role="tool",
                    tool_call_id=tool_call.id,
                    tool_name=tool_name,
                    content=str(result),
                )
        except Exception as e:
            logger.error(f"Error executing tool '{tool_name}': {e}")

        return Message(
            role="tool",
            tool_call_id=tool_call.id,
            tool_name=tool_name,
            content=f"Error executing tool '{tool_name}'",
        )


    def handle_tool_calls(self, tool_calls: List[ChatCompletionMessageToolCall]) -> List:
        
        tool_map = self.get_tool_map()
        if not tool_calls or not tool_map:
            return []
        
        messages = []
        for tool_call in tool_calls:
            observation = self.handle_tool_call(tool_call)
            if observation:
                messages.append(observation.to_dict())
            
        return messages

    async def ahandle_tool_calls(self, tool_calls: List[ChatCompletionMessageToolCall]) -> List:
        
        tool_map = self.get_tool_map()
        if not tool_calls or not tool_map:
            return []

        tasks = []
        for tool_call in tool_calls:
            task = asyncio.create_task(asyncio.to_thread(self.handle_tool_call, tool_call))
            tasks.append(task)

        messages = []

        results = await asyncio.gather(*tasks)
        for observation in results:
            if observation:
                messages.append(observation)

        return messages



import uuid
import json
import asyncio
# import mem0

from typing import List, Union, Optional, Any, Dict, Iterator, Literal
from pydantic import BaseModel, model_validator, field_validator, Field, ConfigDict, UUID4
from datetime import datetime

from gwenflow.llms import ChatBase, ChatOpenAI, ModelResponse
from gwenflow.types import Message
from gwenflow.tools import BaseTool
from gwenflow.memory import ChatMemoryBuffer
from gwenflow.retriever import Retriever
from gwenflow.agents.types import AgentResponse
from gwenflow.agents.prompts import PROMPT_ROLE, PROMPT_STEPS, PROMPT_INSTRUCTIONS, PROMPT_JSON_SCHEMA, PROMPT_CONTEXT, PROMPT_KNOWLEDGE
from gwenflow.utils import logger
from gwenflow.types import Message


class Agent(BaseModel):

    id: UUID4 = Field(default_factory=uuid.uuid4, frozen=True)
    """The unique id of the agent."""

    name: str
    """The name of the agent."""

    description: str | None = None
    """A description of the agent, used as a handoff, so that the manager knows what it does."""

    system_prompt: str | None = None
    """Systtem prompt."""

    instructions: (str | List[str] | None) = None
    """The instructions for the agent."""

    response_model: Dict | None = None
    """Response model."""

    llm: Optional[ChatBase] = Field(None, validate_default=True)
    """The model implementation to use when invoking the LLM."""

    tools: List[BaseTool] | None = None
    """A list of tools that the agent can use."""
    
    tool_choice: Literal["auto", "required", "none"] | str | None = None
    """The tool choice to use when calling the model."""

    reasoning_model: Optional[Any] = Field(None, validate_default=True)
    """Reasoning model."""

    reasoning_steps: str | None = None
    """Reasoning steps."""

    history: ChatMemoryBuffer | None = None
    """Historcal messages for the agent."""

    retriever: Optional[Retriever] = None
    """Retriever for the agent."""

    team: List["Agent"] | None = None
    """Team of agents."""

    model_config = ConfigDict(arbitrary_types_allowed=True)

    @field_validator("id", mode="before")
    @classmethod
    def deny_user_set_id(cls, v: Optional[UUID4]) -> None:
        if v:
            raise ValueError("This field is not to be set by the user.")

    @field_validator("instructions", mode="before")
    @classmethod
    def set_instructions(cls, v: Optional[Union[List, str]]) -> str:
        if isinstance(v, str):
            instructions = [v]
            return instructions
        return v

    @field_validator("llm", mode="before")
    @classmethod
    def set_llm(cls, v: Optional[Any]) -> Any:
        llm = v or ChatOpenAI(model="gpt-4o-mini")
        return llm

    @model_validator(mode="after")
    def model_valid(self) -> Any:
        if self.llm:
            if self.history is None:
                token_limit = self.llm.get_context_window_size()
                self.history = ChatMemoryBuffer(token_limit=token_limit)
            if self.response_model:
                self.llm.response_format = {"type": "json_object"}
            if self.tools:
                self.llm.tools = self.tools
                self.llm.tool_choice = self.tool_choice
        # if self.short_term_memory is None:
        #      self.short_term_memory = mem0.Memory()
        return self

    def _format_context(self, context: Optional[Union[str, Dict[str, str]]]) -> str:
        text = ""
        if isinstance(context, str):
            text = f"<context>\n{ context }\n</context>\n\n"
        elif isinstance(context, dict):
            for key in context.keys():
                text += f"<{key}>\n"
                text += context.get(key) + "\n"
                text += f"</{key}>\n\n"
        return text
    
    def get_system_prompt(self, task: str, context: Optional[Union[str, Dict[str, str]]] = None,) -> str:
        """Get the system prompt for the agent."""

        if self.system_prompt:
            prompt = self.system_prompt
            prompt += "\n\n"
        
        else:
            prompt = PROMPT_ROLE.format(name=self.name, date=datetime.now()).strip()
            prompt += "\n\n"

            # instructions
            instructions = []
            if self.response_model:
                instructions.append("Use JSON to format your answers.")    
            if context is not None:
                instructions.append("Always prefer information from the provided context over your own knowledge.")
            if self.instructions:
                if isinstance(self.instructions, str):
                    instructions += [self.instructions]
                elif isinstance(self.instructions, list):
                    instructions += self.instructions
            if len(instructions) > 0:
                instructions = "\n".join([f"- {i}" for i in instructions])
                prompt += PROMPT_INSTRUCTIONS.format(instructions=instructions).strip()
                prompt += "\n\n"

        # reasoning steps
        if self.reasoning_steps:
            prompt += PROMPT_STEPS.format(reasoning_steps=self.reasoning_steps).strip()
            prompt += "\n\n"

        # response model
        if self.response_model:
            prompt += PROMPT_JSON_SCHEMA.format(json_schema=json.dumps(self.response_model, indent=4)).strip()
            prompt += "\n\n"

        # references
        if self.retriever:
            references = self.retriever.search(query=task)
            if len(references)>0:
                references = [r.content for r in references]
                prompt += PROMPT_KNOWLEDGE.format(references="\n\n".join(references)).strip()
                prompt += "\n\n"

        # context
        if context is not None:
            prompt += PROMPT_CONTEXT.format(context=self._format_context(context)).strip()
            prompt += "\n\n"

        return prompt.strip()
    
    # def reason(self, task: str):

    #     if self.reasoning_model is None:
    #         return None
        
    #     user_prompt = ""
        
    #     if self.tools:
    #         tools = self.get_tools_text_schema()
    #         user_prompt += PROMPT_TOOLS.format(tools=tools).strip() + "\n\n"

    #     user_prompt += PROMPT_TASK.format(task=task).strip()
    #     user_prompt += "\n\nPlease help me with some thoughts, steps and guidelines to answer accurately and precisely to this task."

    #     params = {
    #         "messages": [{"role": "user", "content": user_prompt}],
    #     }

    #     logger.debug("Reasoning.")
    #     response = self.reasoning_model.invoke(**params)

    #     # only keep text outside <think>
    #     reasoning_content = re.sub(r'<think>.*?</think>', '', response, flags=re.DOTALL)
    #     if not reasoning_content:
    #         return None
        
    #     reasoning_content = reasoning_content.strip()
        
    #     logger.debug("Thought: " + reasoning_content)

    #     return reasoning_content
    
    def _run(
        self,
        messages: Union[str, List[Message], List[Dict[str, str]]],
        context: Optional[Union[str, Dict[str, str]]] = None,
        stream: Optional[bool] = False,
    ) ->  Iterator[AgentResponse]:

        messages = self.llm._cast_messages(messages)

        # init agent and model response
        agent_response = AgentResponse()
        model_response = ModelResponse()

        # add reasoning steps
        # if self.reasoning_model:
        #     tools = self.get_tools_text_schema()
        #     completion = self.invoke(PROMPT_REASONING_STEPS_TOOLS.format(task=task, tools=tools))
        #     if len(completion.choices)>0:
        #         self.reasoning_steps = completion.choices[0].message.content

        # search for references
        task = messages[-1].content

        # history
        self.history.system_prompt = self.get_system_prompt(task=task, context=context)
        self.history.add_messages(messages)

        # format messages
        messages_for_model = [m.to_dict() for m in self.history.get()]

        # call llm and tool
        if stream:
            for chunk in self.llm.response_stream(messages=messages_for_model):
                agent_response.content = None
                agent_response.thinking = None
                if chunk.content:
                    model_response.content += chunk.content
                    agent_response.content  = chunk.content
                if chunk.thinking:
                    model_response.thinking += chunk.thinking
                    agent_response.thinking  = chunk.thinking
                yield agent_response
        
        else:
            model_response = self.llm.response(messages=messages_for_model)
            agent_response.content = model_response.content

        # add messages to the current message stack
        self.history.add_message(Message(role="assistant", content=model_response.content))

        # format response
        if self.response_model:
            agent_response.content = json.loads(agent_response.content)

        agent_response.finish_reason = "stop"

        yield agent_response


    def run(
        self,
        messages: Union[str, List[Message], List[Dict[str, str]]],
        context: Optional[Union[str, Dict[str, str]]] = None,
        stream: Optional[bool] = False,
    ) ->  Union[AgentResponse, Iterator[AgentResponse]]:

        logger.debug("")
        logger.debug("------------------------------------------")
        logger.debug(f"Running Agent: { self.name }")
        logger.debug("------------------------------------------")
        logger.debug("")

        if stream:
            response = self._run(messages=messages, context=context, stream=True)
            return response
    
        response = self._run(messages=messages, context=context, stream=False)
        return next(response)




import uuid
import json
import inspect
import asyncio

from typing import List, Union, Optional, Any, Dict, Iterator
from collections import defaultdict
from pydantic import BaseModel, model_validator, field_validator, Field, UUID4
from datetime import datetime

from gwenflow.llms import ChatBase, ChatOpenAI
from gwenflow.types import Message, ChatCompletionChunk
from gwenflow.tools import BaseTool
from gwenflow.memory import ChatMemoryBuffer
from gwenflow.agents import Agent
from gwenflow.agents.types import AgentResponse
from gwenflow.utils import logger
from gwenflow.types import Message


PROMPT_REFORMULATE = """\
# Instructions
- You are an AI assistant reading a current user query and chat_history.
- Given the chat_history, and current user's query, infer the user's intent expressed in the current user query.
- Once you infer the intent, respond with a task that can be used to with an Agent for the current user's query based on the intent
- Be specific in what the user is asking about, but disregard parts of the chat history that are not relevant to the user's intent.
- Provide responses in JSON format

# Examples
Example 1:
With a conversation like below:
```
 - user: are the trailwalker shoes waterproof?
 - assistant: Yes, the TrailWalker Hiking Shoes are waterproof. They are designed with a durable and waterproof construction to withstand various terrains and weather conditions.
 - user: how much do they cost?
```
Respond with:
{
    "intent": "The user wants to know how much the Trailwalker Hiking Shoes cost.",
    "task": "price of Trailwalker Hiking Shoes"
}

Example 2:
With a conversation like below:
```
 - user: are the trailwalker shoes waterproof?
 - assistant: Yes, the TrailWalker Hiking Shoes are waterproof. They are designed with a durable and waterproof construction to withstand various terrains and weather conditions.
 - user: how much do they cost?
 - assistant: The TrailWalker Hiking Shoes are priced at $110.
 - user: do you have waterproof tents?
 - assistant: Yes, we have waterproof tents available. Can you please provide more information about the type or size of tent you are looking for?
 - user: which is your most waterproof tent?
 - assistant: Our most waterproof tent is the Alpine Explorer Tent. It is designed with a waterproof material and has a rainfly with a waterproof rating of 3000mm. This tent provides reliable protection against rain and moisture.
 - user: how much does it cost?
```
Respond with:
{
    "intent": "The user would like to know how much the Alpine Explorer Tent costs.",
    "task": "price of Alpine Explorer Tent"
}

Return the task for the messages in the following conversation:

"""

class ChatAgent(BaseModel):

    llm: Optional[ChatBase] = Field(None, validate_default=True)
    agent: Agent

    @field_validator("llm", mode="before")
    @classmethod
    def set_llm(cls, v: Optional[Any]) -> Any:
        llm = v or ChatOpenAI(model="gpt-4o-mini")
        return llm
    
    def run(self, messages: Union[str, List[Message], List[Dict[str, str]]]) -> Any:

        messages = self.llm._cast_messages(messages)

        conversation = []
        for message in messages:
            conversation.append(f"- {message.role}: {message.content}")
        conversation = "\n".join(conversation)
        prompt = PROMPT_REFORMULATE + f"<conversation>\n{conversation}\n</conversation>"

        self.llm.response_format = {"type": "json_object"}
        response = self.llm.invoke(prompt)
        response = response.choices[0].message.content
        print("Task:", response["task"])

        # if response["agentic"] is False:

        return self.agent.run(response["task"])



import os
import logging
import hashlib
import numpy as np
import pickle
from typing import Optional, Any

try:
    import faiss
except ImportError:
    raise ImportError("`faiss` is not installed.")

from gwenflow.vector_stores.base import VectorStoreBase
from gwenflow.embeddings import Embeddings, GwenlakeEmbeddings
from gwenflow.reranker import Reranker
from gwenflow.types import Document


logger = logging.getLogger(__name__)


class FAISS(VectorStoreBase):

    def __init__(
        self,
        filename: str,
        embeddings: Embeddings = GwenlakeEmbeddings(),
        reranker: Optional[Reranker] = None,
    ):

        # Embedder
        self.embeddings = embeddings

        # reranker
        self.reranker = reranker

        # name
        self.filename = filename
        if not self.filename.endswith(".pkl"):
            self.filename = self.filename + ".pkl"

        self.index = None
        self.metadata = []
        self.create()

    def create(self):
        if not self.exists():
            self.index = faiss.IndexFlatL2(self.embeddings.dimensions)
            self.metadata = []
            self.save()
        else:
            self.load()

    def exists(self) -> bool:
        if os.path.isfile(self.filename):
            return True
        return False
    
    def get_collections(self) -> list:
        return []

    def insert(self, documents: list[Document]):
        logger.info(f"Embedding {len(documents)} documents")
        embeddings = self.embeddings.embed_documents([document.content for document in documents])
        embeddings = np.array(embeddings, dtype='float32')

        logger.info(f"Inserting {len(documents)} documents into index")
        data = []
        for document in documents:
            if document.id is None:
                document.id = hashlib.md5(document.content.encode(), usedforsecurity=False).hexdigest()
            data.append(document.model_dump())
    
        if len(documents) > 0:
            self.index.add(embeddings)
            self.metadata.extend(data)
            self.save()

    def search(self, query: str, limit: int = 5) -> list[Document]:

        if not self.index:
            logger.error(f"Error no index.")
            return []

        query_embedding = self.embeddings.embed_query(query)
        if query_embedding is None:
            logger.error(f"Error getting embedding for Query: {query}")
            return []
        query_embedding = np.array([query_embedding], dtype='float32')
        
        D, I = self.index.search(query_embedding, k=limit)

        documents = []
        for idx, score in zip(I[0], D[0]):
            document = self.metadata[idx].copy()
            document.pop("embedding")
            document = Document(**document)
            document.score = score
            documents.append(document)
    
        if self.reranker:
            documents = self.reranker.rerank(query=query, documents=documents)

        return documents

    def save(self):
        try:
            faiss_data = dict(index=self.index, metadata=self.metadata)
            with open(self.filename, "wb") as f:
                pickle.dump(faiss_data, f)
            return True
        except Exception as e:
            logger.error(e)
        return False

    def load(self):
        try:
            with open(self.filename, "rb") as f:
                faiss_data = pickle.load(f)
                self.index = faiss_data["index"]
                self.metadata = faiss_data["metadata"]
        except Exception as e:
            logger.error(e)
    
    def drop(self):
        try:
            os.remove(self.filename)
            self.index = faiss.IndexFlatL2(self.embeddings.dimensions)
            self.metadata = []
            return True
        except Exception as e:
            logger.error(e)
        return False

    def count(self) -> int:
        return 0

    def info(self) -> dict:
        return {}
    
    def delete(self, id: int):
        return False

    def get(self, id: int) -> dict:        
        return None

    def list(self, filters: dict = None, limit: int = 100) -> list:
        return []



import importlib.metadata

try:
    __version__ = importlib.metadata.version(__package__)
except importlib.metadata.PackageNotFoundError:
    # Case where package metadata is not available.
    __version__ = ""


from gwenflow.llms import ChatGwenlake, ChatOpenAI, ChatAzureOpenAI, ChatOllama
from gwenflow.readers import SimpleDirectoryReader
from gwenflow.agents import Agent, ChatAgent
from gwenflow.tools import Tool
from gwenflow.flows import Flow, AutoFlow
from gwenflow.types import Document, Message
from gwenflow.retriever import Retriever


__all__ = [
    "ChatGwenlake",
    "ChatOpenAI",
    "ChatAzureOpenAI",
    "ChatOllama",
    "Document",
    "Message",
    "SimpleDirectoryReader",
    "Retriever",
    "Agent",
    "ChatAgent",
    "Tool",
    "Flow",
    "AutoFlow",
]



import uuid
from typing import List, Optional, Any
from pydantic import BaseModel, Field, field_validator, UUID4
from time import time


class ModelResponse(BaseModel):
    id: UUID4 = Field(default_factory=uuid.uuid4, frozen=True)
    content: Optional[str] = ""
    thinking: Optional[str] = ""
    finish_reason: Optional[str] = None
    created_at: int = Field(default_factory=lambda: int(time()))

    @field_validator("id", mode="before")
    @classmethod
    def deny_user_set_id(cls, v: Optional[UUID4]) -> None:
        if v:
            raise ValueError("This field is not to be set by the user.")



import dirtyjson
import json
import logging
import os
from typing import Optional, Union, Any, List, Dict, Iterator

from gwenflow.llms import ChatBase
from gwenflow.llms.response import ModelResponse
from gwenflow.types import Message, ChatCompletion, ChatCompletionChunk
from gwenflow.utils import extract_json_str
from openai import OpenAI, AsyncOpenAI

logger = logging.getLogger(__name__)


class ChatOpenAI(ChatBase):
 
    model: str = "gpt-4o-mini"

    # model parameters
    temperature: Optional[float] = None
    top_p: Optional[float] = None
    n: Optional[int] = None
    stop: Optional[Union[str, List[str]]] = None
    max_completion_tokens: Optional[int] = None
    max_tokens: Optional[int] = None
    presence_penalty: Optional[float] = None
    frequency_penalty: Optional[float] = None
    logit_bias: Optional[Dict[int, float]] = None
    response_format: Optional[Dict[str, Any]] = None
    seed: Optional[int] = None
    logprobs: Optional[bool] = None
    top_logprobs: Optional[int] = None

    # clients
    client: Optional[OpenAI] = None
    async_client: Optional[AsyncOpenAI] = None

    # client parameters
    api_key: Optional[str] = None
    organization: Optional[str] = None
    base_url: Optional[str] = None
    timeout: Optional[Union[float, int]] = None
    max_retries: Optional[int] = None

    def _get_client_params(self) -> Dict[str, Any]:

        api_key = self.api_key
        if api_key is None:
            api_key = os.environ.get("OPENAI_API_KEY")
        if api_key is None:
            raise ValueError(
                "The api_key client option must be set either by passing api_key to the client or by setting the OPENAI_API_KEY environment variable"
            )

        organization = self.organization
        if organization is None:
            organization = os.environ.get('OPENAI_ORG_ID')

        client_params = {
            "api_key": api_key,
            "organization": organization,
            "base_url": self.base_url,
            "timeout": self.timeout,
            "max_retries": self.max_retries,
        }

        client_params = {k: v for k, v in client_params.items() if v is not None}

        return client_params

    @property
    def _model_params(self) -> Dict[str, Any]:

        model_params = {
            "temperature": self.temperature,
            "top_p": self.top_p,
            "n": self.n,
            "stop": self.stop,
            "max_tokens": self.max_tokens or self.max_completion_tokens,
            "presence_penalty": self.presence_penalty,
            "frequency_penalty": self.frequency_penalty,
            "logit_bias": self.logit_bias,
            "response_format": self.response_format,
            "seed": self.seed,
            "logprobs": self.logprobs,
            "top_logprobs": self.top_logprobs,
        }

        if self.tools:
            openai_schema = [tool.openai_schema for tool in self.tools]
            model_params["tools"] = openai_schema or None
            model_params["tool_choice"] = self.tool_choice or "auto"
        
        model_params = {k: v for k, v in model_params.items() if v is not None}

        return model_params
    
    def get_client(self) -> OpenAI:

        if self.client:
            return self.client
        
        client_params = self._get_client_params()

        self.client = OpenAI(**client_params)
        return self.client

    def get_async_client(self) -> AsyncOpenAI:

        if self.client:
            return self.client
        
        client_params = self._get_client_params()

        self.async_client = AsyncOpenAI(**client_params)
        return self.async_client

    def _parse_response(self, response: str, response_format: dict = None) -> str:
        """Process the response."""

        if response_format.get("type") == "json_object":
            try:
                json_str = extract_json_str(response)
                # text_response = dirtyjson.loads(json_str)
                text_response = json.loads(json_str)
                return text_response
            except:
                pass

        return response

    def _format_message(self, message: Message) -> Dict[str, Any]:
        """Format a message into the format expected by OpenAI."""

        message_dict: Dict[str, Any] = {
            "role": message.role,
            "content": message.content,
            "name": message.name,
            "tool_call_id": message.tool_call_id,
            "tool_calls": message.tool_calls,
        }
        message_dict = {k: v for k, v in message_dict.items() if v is not None}

        # OpenAI expects the tool_calls to be None if empty, not an empty list
        if message.tool_calls is not None and len(message.tool_calls) == 0:
            message_dict["tool_calls"] = None

        # Manually add the content field even if it is None
        if message.content is None:
            message_dict["content"] = None

        return message_dict

    def _get_thinking(self, tool_calls) -> str:
        thinking = []
        for tool_call in tool_calls:
            if not isinstance(tool_call, dict):
                tool_call = tool_call.model_dump()
            arguments = json.loads(tool_call["function"]["arguments"])
            arguments = ", ".join(arguments.values())
            thinking.append(f"""**Calling** { tool_call["function"]["name"].replace("Tool","") } on '{ arguments }'""")
        if len(thinking)>0:
            return "\n".join(thinking)
        return ""
    
    def invoke(self, messages: Union[str, List[Message], List[Dict[str, str]]]) -> ChatCompletion:
        messages_for_model = self._cast_messages(messages)

        try:
                    
            completion = self.get_client().chat.completions.create(
                model=self.model,
                messages=[self._format_message(m) for m in messages_for_model],
                **self._model_params,
            )

        except Exception as e:
            raise RuntimeError(f"Error in calling openai API: {e}")

        completion = ChatCompletion(**completion.model_dump())

        if self.response_format:
            completion.choices[0].message.content = self._parse_response(completion.choices[0].message.content, response_format=self.response_format)

        return completion

    async def ainvoke(self, messages: Union[str, List[Message], List[Dict[str, str]]]) -> ChatCompletion:
        messages_for_model = self._cast_messages(messages)

        try:

            completion = await self.get_async_client().chat.completions.create(
                model=self.model,
                messages=[self._format_message(m) for m in messages_for_model],
                **self._model_params,
            )
        except Exception as e:
            raise RuntimeError(f"Error in calling openai API: {e}")

        completion = ChatCompletion(**completion.model_dump())

        if self.response_format:
            completion.choices[0].message.content = self._parse_response(completion.choices[0].message.content, response_format=self.response_format)

        return completion

    def stream(self, messages: Union[str, List[Message], List[Dict[str, str]]]) -> Iterator[ChatCompletionChunk]:
        messages_for_model = self._cast_messages(messages)

        try:
            completion = self.get_client().chat.completions.create(
                model=self.model,
                messages=[self._format_message(m) for m in messages_for_model],
                stream=True,
                stream_options={"include_usage": True},
                **self._model_params,
            )
        except Exception as e:
            raise RuntimeError(f"Error in calling openai API: {e}")

        for chunk in completion:
            chunk = ChatCompletionChunk(**chunk.model_dump())
            yield chunk

    async def astream(self, messages: Union[str, List[Message], List[Dict[str, str]]]) ->  Any:
        messages_for_model = self._cast_messages(messages)

        try:

            completion = await self.get_client().chat.completions.create(
                model=self.model,
                messages=[self._format_message(m) for m in messages_for_model],
                stream=True,
                stream_options={"include_usage": True},
                **self._model_params,
            )
        except Exception as e:
            raise RuntimeError(f"Error in calling openai API: {e}")

        async for chunk in completion:
            chunk = ChatCompletionChunk(**chunk.model_dump())
            yield chunk

    def response(self, messages: Union[str, List[Message], List[Dict[str, str]]]) -> ModelResponse:
        messages_for_model = self._cast_messages(messages)
        model_response = ModelResponse()

        while True:

            response = self.invoke(messages=messages_for_model)
            
            if not response.choices[0].message.tool_calls:
                model_response.content = response.choices[0].message.content
                break

            model_response.thinking = self._get_thinking(response.choices[0].message.tool_calls)

            tool_calls = response.choices[0].message.tool_calls
            if tool_calls and self.tools:
                tool_messages = self.handle_tool_calls(tool_calls=tool_calls)
                if len(tool_messages)>0:
                    messages_for_model.append(response.choices[0].message.model_dump())
                    messages_for_model.extend(tool_messages)
        
        return model_response

    async def aresponse(self, messages: Union[str, List[Message], List[Dict[str, str]]]) -> ModelResponse:
        messages_for_model = self._cast_messages(messages)
        model_response = ModelResponse()

        while True:

            response = await self.ainvoke(messages=messages_for_model)

            if not response.choices[0].message.tool_calls:
                model_response.content = response.choices[0].message.content
                break

            model_response.thinking = self._get_thinking(response.choices[0].message.tool_calls)

            tool_calls = response.choices[0].message.tool_calls
            if tool_calls and self.tools:
                tool_messages = await self.ahandle_tool_calls(tool_calls=tool_calls)
                if len(tool_messages)>0:
                    messages_for_model.append(response.choices[0].message.model_dump())
                    messages_for_model.extend(tool_messages)

        return model_response

    def response_stream(self, messages: Union[str, List[Message], List[Dict[str, str]]]) -> Iterator[ModelResponse]:
        messages_for_model = self._cast_messages(messages)
        model_response = ModelResponse()

        while True:

            message = Message(role="assistant", content="", delta="", tool_calls=[])

            for chunk in self.stream(messages=messages_for_model):
                chunk = ChatCompletionChunk(**chunk.model_dump())
                if len(chunk.choices) > 0:
                    if chunk.choices[0].delta.content:
                        message.content += chunk.choices[0].delta.content
                        model_response.content = chunk.choices[0].delta.content
                        model_response.thinking = None
                        yield model_response
                    elif chunk.choices[0].delta.tool_calls:
                        if chunk.choices[0].delta.tool_calls[0].id:
                            message.tool_calls.append(chunk.choices[0].delta.tool_calls[0].model_dump())
                        if chunk.choices[0].delta.tool_calls[0].function.arguments:
                            current_tool = len(message.tool_calls) - 1
                            message.tool_calls[current_tool]["function"]["arguments"] += chunk.choices[0].delta.tool_calls[0].function.arguments

            if not message.tool_calls:
                model_response.content = None
                model_response.finish_reason = "stop"
                break

            model_response.thinking = self._get_thinking(message.tool_calls)
            if model_response.thinking:
                yield model_response

            tool_calls = message.tool_calls
            if tool_calls and self.tools:
                tool_messages = self.handle_tool_calls(tool_calls=tool_calls)
                if len(tool_messages)>0:
                    messages_for_model.append(message)
                    messages_for_model.extend(tool_messages)

        yield model_response

    async def aresponse_stream(self, messages: Union[str, List[Message], List[Dict[str, str]]]) -> Any:
        messages_for_model = self._cast_messages(messages)
        model_response = ModelResponse()

        while True:

            model_response.thinking = []

            message = Message(role="assistant", content="", delta="", tool_calls=[])

            for chunk in await self.astream(messages=messages_for_model):
                chunk = ChatCompletionChunk(**chunk.model_dump())
                if len(chunk.choices) > 0:
                    if chunk.choices[0].delta.content:
                        message.content += chunk.choices[0].delta.content
                        model_response.content = chunk.choices[0].delta.content
                        model_response.thinking = None
                        yield model_response
                    elif chunk.choices[0].delta.tool_calls:
                        if chunk.choices[0].delta.tool_calls[0].id:
                            message.tool_calls.append(chunk.choices[0].delta.tool_calls[0].model_dump())
                        if chunk.choices[0].delta.tool_calls[0].function.arguments:
                            current_tool = len(message.tool_calls) - 1
                            message.tool_calls[current_tool]["function"]["arguments"] += chunk.choices[0].delta.tool_calls[0].function.arguments

            if not message.tool_calls:
                model_response.content = None
                model_response.finish_reason = "stop"
                break

            model_response.thinking = self._get_thinking(message.tool_calls)
            if model_response.thinking:
                yield model_response

            tool_calls = message.tool_calls
            if tool_calls and self.tools:
                tool_messages = await self.ahandle_tool_calls(tool_calls=tool_calls)
                if len(tool_messages)>0:
                    messages_for_model.append(message)
                    messages_for_model.extend(tool_messages)

        yield model_response



from typing import Optional
from pydantic import field_validator, Field

from gwenflow.types import Message
from gwenflow.memory.base import BaseChatMemory


DEFAULT_TOKEN_LIMIT = 8192
DEFAULT_TOKEN_LIMIT_RATIO = 0.75


class ChatMemoryBuffer(BaseChatMemory):
 
    token_limit: Optional[int] = Field(None, validate_default=True)

    @field_validator("token_limit", mode="before")
    def set_token_limit(cls, v: Optional[int]) -> int:
        token_limit = v or int(DEFAULT_TOKEN_LIMIT * DEFAULT_TOKEN_LIMIT_RATIO)
        return token_limit
    
    def get(self):

        initial_token_count = 0

        if self.system_prompt:
            initial_token_count = self._token_count_for_messages([Message(role="system", content=self.system_prompt)])
        if initial_token_count > self.token_limit:
            raise ValueError("Initial token count exceeds token limit")

        chat_history = self.messages    
        message_count = len(chat_history)

        cur_messages = chat_history[-message_count:]
        token_count = self._token_count_for_messages(cur_messages) + initial_token_count

        while token_count > self.token_limit and message_count > 1:
            message_count -= 1
            while chat_history[-message_count].role in ("tool", "assistant"):
                message_count -= 1
            cur_messages = chat_history[-message_count:]
            token_count = self._token_count_for_messages(cur_messages) + initial_token_count

        # catch one message longer than token limit
        if token_count > self.token_limit or (message_count <= 0 and not self.system_prompt):
            return []

        if self.system_prompt:
            return [Message(role="system", content=self.system_prompt)] + chat_history[-message_count:]
        
        return chat_history[-message_count:]




from typing import List, Dict, Any, Optional
from pydantic import BaseModel, model_validator, field_validator, Field, ConfigDict

import hashlib
from langchain_text_splitters import RecursiveCharacterTextSplitter

from gwenflow.types.document import Document
from gwenflow.embeddings import GwenlakeEmbeddings
from gwenflow.reranker import GwenlakeReranker
from gwenflow.vector_stores.base import VectorStoreBase
from gwenflow.vector_stores.lancedb import LanceDB
from gwenflow.utils import logger


MIN_CONTENT_LENGTH = 20


class Retriever(BaseModel):

    name: str

    pathname: Optional[str] = None
    vector_db: Optional[VectorStoreBase] = None
    chunk_size: Optional[int] = 500
    chunk_overlap: Optional[int] = 100
    top_k: Optional[int] = 5

    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)
   
    @model_validator(mode="after")
    def model_valid(self) -> Any:
        if not self.vector_db:
            try:
                uri = self.pathname or f"./{ self.name }"
                self.vector_db = LanceDB(uri=uri,
                                         embeddings=GwenlakeEmbeddings(model="multilingual-e5-large"),
                                         reranker=GwenlakeReranker(model="BAAI/bge-reranker-v2-m3"))
            except Exception as e:
                logger.error(f"Error creating retriver: {e}")
        return self
    
    def search(self, query: str, filters: dict = None) -> list[Document]:
        try:
            if not self.vector_db:
                return []
            documents = self.vector_db.search(query, limit=10*self.top_k, filters=filters)
            return documents[:self.top_k]
        except Exception as e:
            logger.error(f"Error searching for documents: {e}")
        return []

    def _unique_key(self, text: str):
        return hashlib.md5(text.encode(), usedforsecurity=False).hexdigest()
    
    def load_document(self, document: Document) -> bool:
        if not self.vector_db:
            return False

        try:
            docs = []
            text_splitter = RecursiveCharacterTextSplitter(chunk_size=self.chunk_size, chunk_overlap=self.chunk_overlap)

            for doc in text_splitter.create_documents([document.content]):
                if len(doc.page_content) > MIN_CONTENT_LENGTH:
                    docs.append(
                        Document(
                            id=self._unique_key(doc.page_content), # id based on content to remove content duplicates in knowledge
                            content=doc.page_content,
                        )
                    )

            if len(docs)>0:
                self.vector_db.insert(docs)

            return True
        
        except Exception as e:
            logger.error(f"Error loading document: {e}")

        return False
    
    def load_documents(self, documents: List[Any]) -> bool:
        for document in documents:
            if isinstance(document, str):
                document = Document(content=document)
            self.load_document(document)
    